@@ -1,23 +1,23 @@
 """AMQP Connections
 """
-import ssl
 import logging
 import socket
 from array import array
 import pprint
-from threading import Event, Thread
+from threading import Event
 from datetime import datetime
+import time
 
-from .proto import Method, Frame
+from . import __version__, compat
+from .proto import Method
 from .method_io import MethodReader, MethodWriter
 from .serialization import AMQPWriter
-from . import __version__
 from .abstract_channel import AbstractChannel
 from .channel import Channel
-from .exceptions import ResourceError, AMQPConnectionError, error_for_code
+from .exceptions import ResourceError, AMQPConnectionError, Timeout, error_for_code
 from .transport import create_transport
 from . import spec
-from .spec import FrameType, method_t
+from .spec import method_t
 from .concurrency import synchronized
 
 __all__ = ['Connection']
@@ -32,6 +32,8 @@ LIBRARY_PROPERTIES = {
 
 log = logging.getLogger('amqpy')
 
+compat.patch()
+
 
 class Connection(AbstractChannel):
     """The connection class provides methods for a client to establish a network connection to a
@@ -81,7 +83,8 @@ class Connection(AbstractChannel):
         # the connection object itself is treated as channel 0
         super().__init__(self, 0)
 
-        ## instance variables
+        # instance variables
+        #: :type: amqpy.transport.Transport
         self.transport = None
         self.method_reader = None
         self.method_writer = None
@@ -116,12 +119,11 @@ class Connection(AbstractChannel):
         self._daemon = daemon
         self._client_properties = client_properties
 
-        ## callbacks
+        # callbacks
         self.on_blocked = on_blocked
         self.on_unblocked = on_unblocked
 
-        ## heartbeat
-        self.last_heartbeat_sent = None
+        # heartbeat
         self._close_event = None  # Event(): the heartbeat thread listens for this event to exit
         self.heartbeat_thread = None
 
@@ -166,15 +168,19 @@ class Connection(AbstractChannel):
 
         # set up automatic heartbeats
         self._close_event = Event()
-        if self._auto_heartbeat:
-            log.debug('Start automatic heartbeat thread')
-            t = Thread(target=self._heartbeat_thread, name='HeartbeatThread', daemon=self._daemon)
-            t.start()
-            self.heartbeat_thread = t
+        # if self._auto_heartbeat:
+        #     log.debug('Start automatic heartbeat thread')
+        #     t = Thread(target=self._heartbeat_thread, name='HeartbeatThread', daemon=self._daemon)
+        #     t.start()
+        #     self.heartbeat_thread = t
 
     @property
     def last_heartbeat_recv(self):
-        return self.method_reader.last_heartbeat_recv
+        return self.transport.last_heartbeat_received
+
+    @property
+    def last_heartbeat_sent(self):
+        return self.transport.last_heartbeat_sent
 
     @property
     def connected(self):
@@ -220,8 +226,7 @@ class Connection(AbstractChannel):
     def send_heartbeat(self):
         """Send a heartbeat to the server
         """
-        self.last_heartbeat_sent = datetime.now()
-        self.transport.write_frame(Frame(FrameType.HEARTBEAT))
+        self.transport.send_heartbeat()
 
     def is_alive(self):
         """Check if connection is alive
@@ -234,37 +239,11 @@ class Connection(AbstractChannel):
         :return: True if connection is alive, else False
         :rtype: bool
         """
-        if not self.sock:
-            # we don't have a valid socket, this connection is definitely not alive
+        if self.transport:
+            return self.transport.is_alive()
+        else:
             return False
 
-        if not self.connected:
-            # the `transport` is not connected
-            return False
-
-        # recv with MSG_PEEK to check if the connection is alive
-        # note: if there is data still in the buffer, this will not tell us anything
-        if hasattr(socket, 'MSG_PEEK') and not isinstance(self.sock, ssl.SSLSocket):
-            prev = self.sock.gettimeout()
-            self.sock.settimeout(0.0001)
-            try:
-                self.sock.recv(1, socket.MSG_PEEK)
-            except socket.timeout:
-                pass
-            except socket.error:
-                # the exception is usually (always?) a ConnectionResetError in Python 3.3+
-                return False
-            finally:
-                self.sock.settimeout(prev)
-
-        # send a heartbeat to check if the connection is alive
-        try:
-            self.send_heartbeat()
-        except socket.error:
-            return False
-
-        return True
-
     def _wait_any(self, timeout=None):
         """Wait for any event on the connection (for any channel)
 
@@ -283,6 +262,38 @@ class Connection(AbstractChannel):
         method = self.method_reader.read_method(timeout)
         return method
 
+    def _wait_any_hb(self, timeout=None):
+        """Wait for any event on the connection (for any channel), while
+        sending heartbeats
+
+        This method serves the same purpose as :meth:`self._wait_any`,
+        but sends regular heartbeats as configured for the connection.
+
+        Note::
+
+            The ``timeout`` parameter only has a +/- 1s accuracy.
+
+        :param float timeout: timeout
+        :return: method
+        :rtype: amqpy.proto.Method
+        :raise amqpy.exceptions.Timeout: if the operation times out
+        """
+        start_time = time.monotonic()
+
+        while True:
+            # send heartbeat if necessary
+            hd = time.monotonic() - self.transport.last_heartbeat_sent_monotonic
+            if hd >= self._heartbeat_final * 0.85:
+                self.send_heartbeat()
+
+            try:
+                return self._wait_any(1)
+            except Timeout:
+                pass
+
+            if timeout and time.monotonic() - start_time >= timeout:
+                raise Timeout()
+
     def drain_events(self, timeout=None):
         """Wait for an event on all channels
 
@@ -293,7 +304,12 @@ class Connection(AbstractChannel):
         :type timeout: float or None
         :raise amqpy.exceptions.Timeout: if the operation times out
         """
-        method = self._wait_any(timeout)
+        if self._heartbeat_final:
+            method = self._wait_any_hb(timeout)
+        else:
+            log.debug('Using self._wait_any()')
+            method = self._wait_any(timeout)
+
         assert isinstance(method, Method)
         channel = self.channels[method.channel_id]
         return self.handle_method(method, channel)
